from sklearn_pandas.pipeline import TransformerPipeline

import ballet.transformer
from ballet.util.typing import OneOrMore, TransformerLike


class EncoderPipeline(TransformerPipeline):
    """Pipeline of target encoder steps

    This wraps sklearn.pipeline.Pipeline. Each step receives a single argument
    ``y`` to their fit and transform methods. This is needed because some
    consumers like MLBlocks passes arguments by keyword, and we need to pass an
    argument named ``y`` rather than one named ``X``.

    Args:
        can_skip_transform_none: behavior if during the transform stage, the
            input y is None (as would be the case during the ``predict`` stage
            of an MLPipeline). If false (the default), then we call the
            pipeline's transform method on y. If true, we skip calling the
            transform method and instead return immediately (returning the
            value ``None``).
    """

    def __init__(self, *args, can_skip_transform_none=False, **kwargs):
        self.can_skip_transform_none = can_skip_transform_none
        super().__init__(*args, **kwargs)

    def fit(self, y, **fit_params):
        return super().fit(X=y, **fit_params)

    def transform(self, y):
        if self.can_skip_transform_none and y is None:
            return None
        return super().transform(X=y)

    def fit_transform(self, y, **fit_params):
        return self.fit(y, **fit_params).transform(y)


def make_encoder_pipeline(steps, **kwargs):
    return EncoderPipeline(
        ballet.transformer._name_estimators(steps), **kwargs)


def make_robust_encoder(
    steps: OneOrMore[TransformerLike],
    **kwargs,
):
    if not isinstance(steps, list):
        steps = [steps]
    steps = [
        ballet.transformer.make_robust_transformer(step)
        for step in steps
    ]
    return make_encoder_pipeline(steps, **kwargs)



import pandas as pd
import pytest

from ballet.encoder import make_robust_encoder
from ballet.eng import IdentityTransformer
from ballet.util.testing import assert_series_equal

with_encoder = pytest.mark.parametrize(
    'encoder',
    [
        None,
        lambda x: x,
        IdentityTransformer(),
        [None],
        [None, None],
    ]
)


@with_encoder
def test_encoder_pipeline(encoder):
    encoder_pipeline = make_robust_encoder(encoder)
    y_df = pd.util.testing.makeFloatSeries()
    encoder_pipeline.fit(y_df)
    y = encoder_pipeline.transform(y_df)
    assert_series_equal(y, y_df)



import funcy as fy
import numpy as np
import pandas as pd
import pytest

from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.pipeline import FeatureEngineeringPipeline

with_input = pytest.mark.parametrize(
    'input',
    [
        ['foo', 'bar'],
        lambda df: ['foo', 'bar'],
    ],
    ids=[
        'list of string',
        'callable to list of string'
    ]
)
with_transformer = pytest.mark.parametrize(
    'transformer',
    [
        IdentityTransformer(),
        [IdentityTransformer()],
        [None, IdentityTransformer(), lambda x: x],
        Feature(['foo', 'bar'], IdentityTransformer()),
        [None, IdentityTransformer(), Feature(
            ['foo', 'bar'], IdentityTransformer())],
    ],
    ids=[
        'scalar',
        'list of transformer',
        'list of mixed',
        'nested feature',
        'list of mixed and nested features',
    ]
)


@with_input
@with_transformer
def test_init(input, transformer):
    feature = Feature(input, transformer)
    mapper = FeatureEngineeringPipeline(feature)
    assert isinstance(mapper, FeatureEngineeringPipeline)


@with_input
@with_transformer
def test_fit(input, transformer):
    feature = Feature(input, transformer)
    mapper = FeatureEngineeringPipeline(feature)
    df = pd.util.testing.makeCustomDataframe(5, 2)
    df.columns = ['foo', 'bar']
    mapper.fit(df)


@with_input
@with_transformer
def test_transform(input, transformer):
    feature = Feature(input, transformer)
    mapper = FeatureEngineeringPipeline(feature)
    df = pd.util.testing.makeCustomDataframe(5, 2)
    df.columns = ['foo', 'bar']
    mapper.fit(df)
    X = mapper.transform(df)
    assert np.shape(X) == (5, 2)


@with_input
@with_transformer
@pytest.mark.parametrize(
    'output',
    [
        None,
        'baz',
        ['foobaz', 'barbaz'],
    ]
)
def test_df_colnames(input, transformer, output):
    feature = Feature(input, transformer, output=output)
    mapper = FeatureEngineeringPipeline(feature)
    entities_df = pd.util.testing.makeCustomDataframe(5, 2)
    entities_df.columns = ['foo', 'bar']
    feature_matrix = mapper.fit_transform(entities_df)
    feature_frame = pd.DataFrame(
        feature_matrix,
        columns=mapper.transformed_names_,
        index=entities_df.index,
    )
    assert fy.all(fy.isa(str), feature_frame.columns)


def test_gh_87():
    X_df = pd.DataFrame(
        data=np.random.rand(5, 4), columns=['A1', 'A2', 'B1', 'B2'])

    def input(df):
        return [col for col in df.columns if col.startswith('A')]

    transformer = None
    feature = Feature(input, transformer)
    X = feature.pipeline.fit_transform(X_df)

    assert X.shape[1] == len(input(X_df))

    X_ft_df = pd.DataFrame(
        X,
        columns=feature.pipeline.transformed_names_,
        index=X_df.index,
    )
    assert all(isinstance(col, str) for col in X_ft_df.columns)



import pathlib
from copy import deepcopy
from typing import Optional

from ballet.encoder import EncoderPipeline, make_robust_encoder
from ballet.pipeline import FeatureEngineeringPipeline
from ballet.project import Project
from ballet.util.mod import import_module_from_modname

PRIMITIVES_PATH = [pathlib.Path(__file__).with_name('primitives').resolve()]
PIPELINES_PATH = [pathlib.Path(__file__).with_name('pipelines').resolve()]

_project_detection_details = """\
The primitive detects the Ballet project in one of three ways. First, if
the package slug is given, then Ballet will import the package and detect
the project. Second, if a path to the project is given (i.e. the directory
containing the ballet.yml configuration file), then the project will be
detected that way. Third, if neither is given, then Ballet will try to
detect a project in the current working directory. This third option is the
least robust and sensitive to where you are running the MLPipeline from.

Args:
    package_slug: name of top-level package for the Ballet project
    project_path: filesystem path to the directory containing a valid
        ballet.yml configuration file
"""


def _get_project(package_slug, project_path):
    if package_slug is not None:
        package = import_module_from_modname(package_slug)
        return Project(package)
    elif project_path is not None:
        return Project.from_path(project_path)
    else:
        return Project.from_cwd()


def make_engineer_features(
    package_slug: Optional[str] = None,
    project_path: Optional[str] = None
) -> FeatureEngineeringPipeline:
    f"""Make the engineer_features primitive for the given Ballet project

    {_project_detection_details}

    Returns:
        a deep copy of the feature engineering pipeline defined by the project
    """
    project = _get_project(package_slug, project_path)
    return deepcopy(project.api.pipeline)


def make_encode_target(
    package_slug: Optional[str] = None,
    project_path: Optional[str] = None
) -> EncoderPipeline:
    f"""Make the encode_target primitive for the given Ballet project

    {_project_detection_details}

    Returns:
        a deep copy of the target encoder pipeline defined by the project
    """
    project = _get_project(package_slug, project_path)
    encoder = deepcopy(project.api.encoder)
    return make_robust_encoder(encoder, can_skip_transform_none=True)
